@@ -22,11 +22,12 @@
 #include <QPointer>
 #include <QUrl>
 #include <curl/curl.h>
-#include "jdnsshared.h"
+#include "qjdnsshared.h"
 #include "bufferlist.h"
 #include "log.h"
 
 #define BUFFER_SIZE 200000
+#define REQUEST_BODY_BUFFER_MAX 1000000
 
 // workaround for earlier curl versions
 #define UNPAUSE_WORKAROUND 1
@@ -44,6 +45,15 @@ static const char *socketActionToString(int x)
 	}
 }
 
+static const char *msgToString(int x)
+{
+	switch(x)
+	{
+		case CURLMSG_DONE: return "CURLMSG_DONE";
+		default: return 0;
+	}
+}
+
 class CurlConnection : public QObject
 {
 	Q_OBJECT
@@ -52,6 +62,7 @@ public:
 	CURLSH *share;
 	CURL *easy;
 	QString method;
+	int maxRedirects;
 	bool expectBody;
 	bool bodyReadFrom;
 	struct curl_slist *dnsCache;
@@ -59,6 +70,7 @@ public:
 	int pauseBits;
 	BufferList in;
 	BufferList out;
+	int outPos;
 	bool inFinished;
 	bool outFinished;
 	bool haveStatusLine;
@@ -72,11 +84,13 @@ public:
 	CURLcode result;
 
 	CurlConnection() :
+		maxRedirects(-1),
 		expectBody(false),
 		bodyReadFrom(false),
 		dnsCache(NULL),
 		headersList(NULL),
 		pauseBits(0),
+		outPos(0),
 		inFinished(false),
 		outFinished(false),
 		haveStatusLine(false),
@@ -100,6 +114,8 @@ public:
 		curl_easy_setopt(easy, CURLOPT_WRITEDATA, this);
 		curl_easy_setopt(easy, CURLOPT_READFUNCTION, readFunction_cb);
 		curl_easy_setopt(easy, CURLOPT_READDATA, this);
+		curl_easy_setopt(easy, CURLOPT_SEEKFUNCTION, seekFunction_cb);
+		curl_easy_setopt(easy, CURLOPT_SEEKDATA, this);
 		curl_easy_setopt(easy, CURLOPT_HEADERFUNCTION, headerFunction_cb);
 		curl_easy_setopt(easy, CURLOPT_HEADERDATA, this);
 
@@ -109,6 +125,10 @@ public:
 
 		if(log_outputLevel() >= LOG_LEVEL_DEBUG)
 			curl_easy_setopt(easy, CURLOPT_VERBOSE, 1);
+
+#if LIBCURL_VERSION_NUM >= 0x072a00
+		curl_easy_setopt(easy, CURLOPT_PATH_AS_IS, 1);
+#endif
 	}
 
 	~CurlConnection()
@@ -126,13 +146,11 @@ public:
 		if(method == "OPTIONS")
 		{
 			expectBody = false;
-			curl_easy_setopt(easy, CURLOPT_NOBODY, 1);
 			curl_easy_setopt(easy, CURLOPT_CUSTOMREQUEST, "OPTIONS");
 		}
 		else if(method == "HEAD")
 		{
 			expectBody = false;
-			curl_easy_setopt(easy, CURLOPT_NOBODY, 1);
 			curl_easy_setopt(easy, CURLOPT_CUSTOMREQUEST, NULL);
 		}
 		else if(method == "GET")
@@ -158,7 +176,6 @@ public:
 		else if(method == "DELETE")
 		{
 			expectBody = false;
-			curl_easy_setopt(easy, CURLOPT_NOBODY, 1);
 			curl_easy_setopt(easy, CURLOPT_CUSTOMREQUEST, "DELETE");
 		}
 		else
@@ -169,7 +186,7 @@ public:
 		}
 	}
 
-	void setup(const QUrl &uri, const HttpHeaders &_headers, const QHostAddress &connectAddr = QHostAddress(), int connectPort = -1)
+	void setup(const QUrl &uri, const HttpHeaders &_headers, const QHostAddress &connectAddr = QHostAddress(), int connectPort = -1, int _maxRedirects = -1)
 	{
 		assert(!method.isEmpty());
 
@@ -222,6 +239,15 @@ public:
 		// disable expect usage as it appears to be buggy
 		curl_slist_append(headersList, "Expect:");
 		curl_easy_setopt(easy, CURLOPT_HTTPHEADER, headersList);
+
+		maxRedirects = _maxRedirects;
+		if(maxRedirects >= 0)
+		{
+			curl_easy_setopt(easy, CURLOPT_FOLLOWLOCATION, 1);
+			curl_easy_setopt(easy, CURLOPT_MAXREDIRS, maxRedirects);
+		}
+
+		curl_easy_setopt(easy, CURLOPT_POSTREDIR, CURL_REDIR_POST_ALL);
 	}
 
 	void update()
@@ -252,6 +278,12 @@ public:
 		return self->readFunction((char *)ptr, size * nmemb);
 	}
 
+	static int seekFunction_cb(void *userdata, curl_off_t offset, int origin)
+	{
+		CurlConnection *self = (CurlConnection *)userdata;
+		return self->seekFunction(offset, origin);
+	}
+
 	static size_t headerFunction_cb(void *ptr, size_t size, size_t nmemb, void *userdata)
 	{
 		CurlConnection *self = (CurlConnection *)userdata;
@@ -298,7 +330,18 @@ public:
 
 	size_t readFunction(char *p, size_t size)
 	{
-		QByteArray buf = out.take(size);
+		QByteArray buf;
+		if(outPos >= 0 && out.size() <= REQUEST_BODY_BUFFER_MAX)
+		{
+			buf = out.mid(outPos, size);
+			outPos += buf.size();
+		}
+		else
+		{
+			outPos = -1; // no longer buffered
+			buf = out.take(size);
+		}
+
 		if(!buf.isEmpty())
 		{
 			bodyReadFrom = true;
@@ -324,6 +367,35 @@ public:
 		}
 	}
 
+	int seekFunction(curl_off_t offset, int origin)
+	{
+		if(outPos < 0)
+		{
+			log_debug("seekFunction: can't seek. input is unbuffered");
+			return 1;
+		}
+
+		if(origin == SEEK_SET)
+		{
+			if(offset <= out.size())
+			{
+				outPos = offset;
+				log_debug("seekFunction: seeking to position %ld", offset);
+				return 0;
+			}
+			else
+			{
+				log_debug("seekFunction: %ld out of range (range: 0-%d)", offset, out.size());
+				return 1;
+			}
+		}
+		else
+		{
+			log_debug("seekFunction: unknown origin value: %d", origin);
+			return 1;
+		}
+	}
+
 	size_t headerFunction(char *p, size_t size)
 	{
 		assert(p[size - 1] == '\n');
@@ -336,7 +408,48 @@ public:
 			len = size - 1;
 
 		QByteArray line(p, len);
-		if(line.isEmpty())
+		if(!line.isEmpty())
+		{
+			if(haveResponseHeaders)
+			{
+				// does it look like we're getting a status
+				//   line again? (happens when redirecting)
+				int at = line.indexOf(' ');
+				if(at != -1 && !line.mid(0, at).contains(':'))
+				{
+					haveStatusLine = false;
+					haveResponseHeaders = false;
+					responseHeaders.clear();
+				}
+			}
+
+			if(!haveResponseHeaders)
+			{
+				if(haveStatusLine)
+				{
+					int at = line.indexOf(": ");
+					if(at == -1)
+						return -1;
+
+					log_debug("response header: %s", line.data());
+					responseHeaders += HttpHeader(line.mid(0, at), line.mid(at + 2));
+				}
+				else
+				{
+					// status reason we have to parse ourselves
+					int at = line.indexOf(' ');
+					if(at == -1)
+						return -1;
+					at = line.indexOf(' ', at + 1);
+					if(at == -1)
+						return -1;
+					responseReason = line.mid(at + 1);
+
+					haveStatusLine = true;
+				}
+			}
+		}
+		else
 		{
 			haveResponseHeaders = true;
 
@@ -350,42 +463,27 @@ public:
 				log_debug("got code 100, ignoring this header block");
 				haveStatusLine = false;
 				haveResponseHeaders = false;
+				responseHeaders.clear();
+				return size;
 			}
-			else
-			{
-				// if a content-encoding was used, don't provide content-length
-				QByteArray contentEncoding = responseHeaders.get("Content-Encoding");
-				if(!contentEncoding.isEmpty() && contentEncoding != "identity")
-					responseHeaders.removeAll("Content-Length");
 
-				// tell the app we've got the header block
-				update();
-			}
-		}
-		else if(!haveResponseHeaders)
-		{
-			if(haveStatusLine)
+			if(maxRedirects >= 0 && responseCode >= 300 && responseCode < 400 && responseHeaders.contains("Location"))
 			{
-				int at = line.indexOf(": ");
-				if(at == -1)
-					return -1;
-
-				log_debug("response header: %s", line.data());
-				responseHeaders += HttpHeader(line.mid(0, at), line.mid(at + 2));
+				log_debug("got code 3xx and redirects enabled, ignoring this header block");
+				haveStatusLine = false;
+				haveResponseHeaders = false;
+				responseHeaders.clear();
+				return size;
 			}
-			else
-			{
-				// status reason we have to parse ourselves
-				int at = line.indexOf(' ');
-				if(at == -1)
-					return -1;
-				at = line.indexOf(' ', at + 1);
-				if(at == -1)
-					return -1;
-				responseReason = line.mid(at + 1);
 
-				haveStatusLine = true;
-			}
+			// if a content-encoding was used, don't provide content-length
+			QByteArray contentEncoding = responseHeaders.get("Content-Encoding");
+			if(!contentEncoding.isEmpty() && contentEncoding != "identity")
+				responseHeaders.removeAll("Content-Length");
+
+			// tell the app we've got the header block
+			newlyReadOrEof = true;
+			update();
 		}
 
 		return size;
@@ -582,14 +680,18 @@ public:
 			if(!m || !m->msg)
 				break;
 
+			const char *str = msgToString(m->msg);
+			if(str)
+				log_debug("message: %s", str);
+			else
+				log_debug("unknown message: %d", m->msg);
+
 			if(m->msg == CURLMSG_DONE)
 			{
 				CurlConnection *conn;
 				curl_easy_getinfo(m->easy_handle, CURLINFO_PRIVATE, &conn);
 				conn->done(m->data.result);
 			}
-			else
-				log_debug("unknown message: %d\n", m->msg);
 		}
 	}
 
@@ -637,9 +739,10 @@ class HttpRequest::Private : public QObject
 
 public:
 	HttpRequest *q;
-	JDnsShared *dns;
+	QJDnsShared *dns;
 	QString connectHost;
 	bool ignoreTlsErrors;
+	int maxRedirects;
 	HttpRequest::ErrorCondition errorCondition;
 	QString method;
 	QUrl uri;
@@ -651,11 +754,13 @@ public:
 	CurlConnection *conn;
 	bool handleAdded;
 
-	Private(HttpRequest *_q, JDnsShared *_dns) :
+	Private(HttpRequest *_q, QJDnsShared *_dns) :
 		QObject(_q),
 		q(_q),
 		dns(_dns),
 		ignoreTlsErrors(false),
+		maxRedirects(-1),
+		errorCondition(HttpRequest::ErrorNone),
 		mostSignificantError(HttpRequest::ErrorGeneric),
 		ignoreBody(false),
 		conn(0),
@@ -824,7 +929,7 @@ public:
 		}
 		else
 		{
-			JDnsSharedRequest *dreq = new JDnsSharedRequest(dns);
+			QJDnsSharedRequest *dreq = new QJDnsSharedRequest(dns);
 			connect(dreq, SIGNAL(resultsReady()), SLOT(dreq_resultsReady()));
 			dreq->query(QUrl::toAce(host), QJDns::A);
 		}
@@ -869,7 +974,7 @@ private slots:
 		if(!self)
 			return;
 
-		conn->setup(uri, headers, addr);
+		conn->setup(uri, headers, addr, -1, maxRedirects);
 
 		if(ignoreTlsErrors)
 		{
@@ -886,7 +991,7 @@ private slots:
 
 	void dreq_resultsReady()
 	{
-		JDnsSharedRequest *dreq = (JDnsSharedRequest *)sender();
+		QJDnsSharedRequest *dreq = (QJDnsSharedRequest *)sender();
 
 		if(dreq->success())
 		{
@@ -930,6 +1035,9 @@ private slots:
 					//   was actually sent off
 					curError = HttpRequest::ErrorTimeout;
 					break;
+				case CURLE_TOO_MANY_REDIRECTS:
+					curError = HttpRequest::ErrorTooManyRedirects;
+					break;
 				default:
 					tryAgain = false;
 					curError = HttpRequest::ErrorGeneric;
@@ -976,7 +1084,7 @@ private slots:
 	}
 };
 
-HttpRequest::HttpRequest(JDnsShared *dns, QObject *parent) :
+HttpRequest::HttpRequest(QJDnsShared *dns, QObject *parent) :
 	QObject(parent)
 {
 	d = new Private(this, dns);
@@ -997,6 +1105,11 @@ void HttpRequest::setIgnoreTlsErrors(bool on)
 	d->ignoreTlsErrors = on;
 }
 
+void HttpRequest::setFollowRedirects(int maxRedirects)
+{
+	d->maxRedirects = maxRedirects;
+}
+
 void HttpRequest::start(const QString &method, const QUrl &uri, const HttpHeaders &headers)
 {
 	d->start(method, uri, headers);
@@ -1022,8 +1135,8 @@ int HttpRequest::bytesAvailable() const
 
 bool HttpRequest::isFinished() const
 {
-	if(d->conn)
-		return d->conn->inFinished;
+	if(d->errorCondition != ErrorNone || (d->conn && d->conn->inFinished))
+		return true;
 	else
 		return false;
 }
